package org.mesopotamia.lang.java;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.mesopotamia.AstSourceUnit;
import org.mesopotamia.LanguageElement;
import org.mesopotamia.LanguageElementHandle;
import org.mesopotamia.MesopotamiaException;
import org.mesopotamia.MesopotamiaToken;
import org.mesopotamia.RepositoryLanguage;
import org.mesopotamia.Scan;
import org.mesopotamia.lang.java.ref.MethodInfo;
import org.mesopotamia.lang.java.ref.Scope;
import org.mesopotamia.lang.java.ref.SoftCachingScope;
import org.mesopotamia.lang.java.ref.TypeInfo;
import org.mesopotamia.lang.java.ref.VariableInfo;
import org.mesopotamia.lang.java.ref.bcel.JavaClassTypeInfoRepository;

import biz.hammurapi.config.Context;
import biz.hammurapi.render.HtmlRenderer;
import biz.hammurapi.render.RenderingException;
import biz.hammurapi.util.Escaper;

public class JavaSourceUnit extends AstSourceUnit implements Scope {
	
	/**
	 * Scan attribute of type ClassLoader for resolution of types. 
	 */
	public static final String SCAN_CLASSLOADER = "SCAN_CLASSLOADER";

	/**
	 * Scan attribute of type ClassLoaderRepository for resolution of types.
	 */
	public static final String SCAN_TYPE_INFO_REPOSITORY = "SCAN_TYPE_INFO_REPOSITORY";
	
	public JavaSourceUnit(
			org.mesopotamia.sql.SourceUnit dbData, 
			Scan scan,
			RepositoryLanguage repoLanguage, 
			Collection<Number> loadLevels)
			throws MesopotamiaException {
		
		super(dbData, scan, repoLanguage, loadLevels);
		synchronized (scan) {
			scanTypeInfoRepository = (JavaClassTypeInfoRepository) scan.getAttribute(SCAN_TYPE_INFO_REPOSITORY);
			if (scanTypeInfoRepository == null) {
				Object scl = scan.getAttribute(SCAN_CLASSLOADER);
				if (scl instanceof ClassLoader) {
					scanTypeInfoRepository = new JavaClassTypeInfoRepository((ClassLoader) scl);
				} else {
					scanTypeInfoRepository = new JavaClassTypeInfoRepository(getClass().getClassLoader());
				}
				scan.setAttribute(SCAN_TYPE_INFO_REPOSITORY, scanTypeInfoRepository);
			}
		}
	}
	
	private JavaClassTypeInfoRepository scanTypeInfoRepository;
		
	public String getPackageName() {
		for (Object child: getLanguageElements()) {
			if (child instanceof PackageDefinition) {
				return ((PackageDefinition) child).getPath();
			}
		}
		return "";
	}
	
	private Scope cachingScope = new SoftCachingScope() {

		@Override
		protected MethodInfo findMethodInternal(String name, String[] parameterTypes) {
			return JavaSourceUnit.this.findMethodInternal(name, parameterTypes);
		}

		@Override
		protected TypeInfo findTypeInternal(String name) {
			return JavaSourceUnit.this.findTypeInternal(name);
		}

		@Override
		protected VariableInfo findVariableInternal(String name) {
			return JavaSourceUnit.this.findVariableInternal(name);
		}
		
	};
	
	public MethodInfo findMethod(String name, String[] parameterTypes) {
		return cachingScope.findMethod(name, parameterTypes);
	}

	public TypeInfo findType(String name) {
		return cachingScope.findType(name);
	}

	public VariableInfo findVariable(String name) {
		return cachingScope.findVariable(name);
	}

	/**
	 * Finds method in static imports.
	 */
	private MethodInfo findMethodInternal(String name, String[] parameterTypes) {
		for (Object le: getLanguageElements()) {
			if (le instanceof ImportDefinition) {
				ImportDefinition iDef = (ImportDefinition) le;
				String path = iDef.getPath();
				if (iDef.isStatic()) {
					TypeInfo staticImport = scanTypeInfoRepository.getTypeInfo(path, this);
					if (staticImport!=null) {
						MethodInfo ret = staticImport.findTypeMethod(name, parameterTypes);
						if (ret!=null) {
							return ret;
						}
					}
				}
			}
		}
		return null;
	}
	
	private static Collection<String> fcnVariants(String fcn) {
		StringTokenizer st = new StringTokenizer(fcn, ".");
		List<String> tokens = new ArrayList<String>();
		while (st.hasMoreTokens()) {
			tokens.add(st.nextToken());
		}
		
		Collection<String> ret = new ArrayList<String>();
		for (int i=0; i<tokens.size(); ++i) {
			StringBuffer sb = new StringBuffer();
			Iterator<String> sit = tokens.iterator();
			for (int j=tokens.size()-1; sit.hasNext(); --j) {
				sb.append(sit.next());
				if (sit.hasNext()) {
					sb.append(j<=i ? "$" : ".");
				}
			}
			ret.add(sb.toString());
		}
		return ret;		
	}
	
	/**
	 * Finds type in defined and imported types.
	 */
	private TypeInfo findTypeInternal(String name) {
		if (Null.NULL_FCN.equals(name)) {
			return Null.NULL_TYPE_INFO;
		}
		
		TypeInfo primitive = BuiltInType.getBuiltInType(name);
		if (primitive!=null) {
			return primitive;
		}
		
		synchronized (fcnMap) {
			LanguageElementHandle handle = fcnMap.get(name);
			if (handle!=null) {
				return (TypeInfo) getScan().getLanguageElement(handle);
			}			
		}
		
		int dotIdx = name.indexOf(".");
		if (dotIdx>0) { // FCN?
			TypeInfo ret = scanTypeInfoRepository.getTypeInfo(name, this);
			if (ret!=null) {
				return ret;
			}
		}
		
		PackageDefinition pDef = null;
		for (Object le: getLanguageElements()) {
			if (le instanceof TypeDefinition) {
				TypeDefinition td = (TypeDefinition) le;
				if (name.equals(td.getName())) {
					return td;
				}
			} else if (le instanceof ImportDefinition) {
				ImportDefinition iDef = (ImportDefinition) le;
				String path = iDef.getPath();
				if (iDef.isStatic()) {
					TypeInfo staticImport = scanTypeInfoRepository.getTypeInfo(path, this);
					if (staticImport!=null) {
						TypeInfo ret = staticImport.findNestedType(name);
						if (ret!=null) {
							return ret;
						}
					}
				} else {
					if (dotIdx==-1 && path.endsWith("."+name)) {
						for (String variant: fcnVariants(path)) {
							TypeInfo ret = scanTypeInfoRepository.getTypeInfo(variant, this);
							if (this!=null) {
								return ret;
							}
						}
						return null;
					}
					
					if (dotIdx>0 && path.endsWith("."+name.substring(0, dotIdx))) {
						TypeInfo ret = scanTypeInfoRepository.getTypeInfo(path+"$"+name.substring(dotIdx+1), this);
						if (ret!=null) {
							return ret;
						}						
					}
					
					if (path.endsWith("*")) {
						TypeInfo ret = scanTypeInfoRepository.getTypeInfo(path.substring(0, path.length()-1)+name, this);
						if (ret!=null) {
							return ret;
						}						
					}
				}
			} else if (le instanceof PackageDefinition) {
				pDef = (PackageDefinition) le;
			}
		}
		
		TypeInfo ret = scanTypeInfoRepository.getTypeInfo("java.lang."+name, this);
		if (ret!=null) {
			return ret;
		}
		
		// Same package.
		if (pDef==null) {
			return scanTypeInfoRepository.getTypeInfo(name, this);							
		} else {
			return scanTypeInfoRepository.getTypeInfo(pDef.getPath()+"."+name, this);							
		}		
	}
	
	/**
	 * Source unit doesn't define variables, but static imports do.
	 */
	private VariableInfo findVariableInternal(String name) {
		for (Object le: getLanguageElements()) {
			if (le instanceof ImportDefinition) {
				ImportDefinition iDef = (ImportDefinition) le;
				String path = iDef.getPath();
				if (iDef.isStatic()) {
					TypeInfo staticImport = scanTypeInfoRepository.getTypeInfo(path, this);
					if (staticImport!=null) {
						VariableInfo ret = staticImport.findTypeVariable(name);
						if (ret!=null) {
							return ret;
						}
					}
				}
			}
		}
		return null;
	}
	
	private Map<String, LanguageElementHandle> fcnMap = new HashMap<String, LanguageElementHandle>();

	public void registerLocalOrAnonymous(String fcn, TypeDefinitionOrAnonymous typeDefinitionOrAnonymous) {
		synchronized (fcnMap) {
			fcnMap.put(fcn, ((LanguageElement) typeDefinitionOrAnonymous).getHandle());			
		}
	}
	
	@Override
	public HtmlRenderer getRenderer(Context context) {
		return new HtmlRenderer() {

			public void render(Writer out) throws IOException, RenderingException {
				out.write("<PRE style=\"border: 1px solid gray\">");
				Collection tokens = (Collection) getLevelData("token");
				if (tokens!=null) {
					java.util.Set<String> constants = new java.util.HashSet<String>();
					constants.add("NUM_INT");
					constants.add("CHAR_LITERAL");
					constants.add("STRING_LITERAL");
					constants.add("NUM_FLOAT");
					constants.add("NUM_LONG");
					constants.add("NUM_DOUBLE");
					constants.add("HEX_DIGIT");
					Iterator it = tokens.iterator();
					int prevLine=-1;
					while (it.hasNext()) {
						MesopotamiaToken token = (MesopotamiaToken) it.next();		
						if (token.getLine()!=prevLine) {
							if (prevLine!=-1) {
								out.write("<BR/>");
							}
							out.write("<a name=\"line_"+token.getLine()+"\" class=\"lineNum\">"+token.getLine()+"</a>");
							prevLine=token.getLine();
						}
						if (token.getText()!=null && !"NEW_LINE".equals(token.getTypeName())) {
							if ("ML_COMMENT".equals(token.getTypeName())) {
								StringTokenizer st = new StringTokenizer(token.getText(), "\n", true);
								prevLine = token.getLine();
								while (st.hasMoreTokens()) {
									String nextToken = st.nextToken();
									if ("\n".equals(nextToken)) {
										++prevLine;
										out.write("<BR/><a name=\"line_"+prevLine+"\" class=\"lineNum\">"+prevLine+"</a>");					
									} else {
										String cLine = ("!"+nextToken).trim().substring(1);
										out.write("<span class=\"comment\">"+Escaper.escapeHtml(cLine)+"</span>");
									}
								}
							} else if ("SL_COMMENT".equals(token.getTypeName())) {
								out.write("<span class=\"comment\">"+("!"+Escaper.escapeHtml(token.getText())).trim().substring(1)+"</span>"); // Trim end.
							} else if (token.getTypeName().startsWith("LITERAL_")) {
								out.write("<span class=\"reserved_word\"><B>"+Escaper.escapeHtml(token.getText())+"</B></span>");
							} else if (constants.contains(token.getTypeName())) {
								out.write("<span class=\"literal\">"+Escaper.escapeHtml(token.getText())+"</span>");
							} else {
								out.write(Escaper.escapeHtml(token.getText()));
							}
						}
					}
				}
				out.write("</PRE>");
			}
			
		};
	}
	
}
