package org.mesopotamia.lang.java;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

import org.mesopotamia.LanguageElement;
import org.mesopotamia.MesopotamiaException;
import org.mesopotamia.MesopotamiaNode;
import org.mesopotamia.NodeData;
import org.mesopotamia.RepositoryLanguage;
import org.mesopotamia.Scan;
import org.mesopotamia.SourceUnit;
import org.mesopotamia.lang.java.TypeDefinition.ParameterMatch;
import org.mesopotamia.lang.java.ref.Info;
import org.mesopotamia.lang.java.ref.MethodInfo;
import org.mesopotamia.lang.java.ref.Reference;
import org.mesopotamia.lang.java.ref.Scope;
import org.mesopotamia.lang.java.ref.TypeInfo;
import org.mesopotamia.lang.java.ref.TypeSpecInfo;
import org.mesopotamia.lang.java.ref.VariableInfo;
import org.w3c.dom.Element;

import biz.hammurapi.util.CollectionVisitable;
import biz.hammurapi.util.Visitor;

@SuppressWarnings("unchecked")
public class NewExpression extends PrimaryExpression implements TypeDefinitionOrAnonymous, Reference<MethodInfo> {
	private static final Logger logger = Logger.getLogger(NewExpression.class.getName());

	private Type specType;
	private int dimensions;
	
	private List<ArrayInitializer> ArrayInitializers;

	public NewExpression(
			NodeData xData, 
			Class<?> context, 
			Scan scan,
			RepositoryLanguage language, 
			Object environment) throws MesopotamiaException {
		super(xData, context, scan, language, environment);

		Arguments = select(Expression.class, "ELIST/*");
//		TypeSpecification = (TypeSpecification) selectSingleElement(TypeSpecification.class, "TYPE");
		Fields = select(Field.class, "OBJBLOCK/*");
		
		LanguageElement firstChild = getChild(0);
		if (firstChild instanceof Type) {
			specType = (Type) firstChild;	
//			if (language.supportsTokenName("TYPE_ARGUMENTS") && specType instanceof LanguageElement) {
//				TypeArguments = ((LanguageElement) specType).select(TypeArgument.class, "TYPE_ARGUMENTS/TYPE_ARGUMENT");
//			}			
			MesopotamiaNode fNode = firstChild.getNode();
			List<MesopotamiaNode> siblings = fNode.getParent().getChildren();
			int nextSiblingPosition = fNode.getPosition()+1;
			if (siblings.size()>nextSiblingPosition) {
				MesopotamiaNode nextSibling = siblings.get(nextSiblingPosition);
				while ("ARRAY_DECLARATOR".equals(nextSibling.getTypeName())) {
					++dimensions;
					siblings = nextSibling.getChildren();
					if (siblings.isEmpty()) {
						break;
					}
					nextSibling = siblings.get(0);
				}
			}
		}
		
		ArrayInitializers = select(ArrayInitializer.class, "ARRAY_INIT/*");
	}

	public void toDom(Element holder) {
		super.toDom(holder);

		// Serialize attributes
//		setElement(holder, "TypeArguments", TypeArguments);
		setElement(holder, "Arguments", Arguments);
//		setElement(holder, "TypeSpecification", TypeSpecification);
	}

	// Attributes
//	private List<TypeArgument> TypeArguments = emptyList;

	private List<Expression> Arguments;

//	private TypeSpecification TypeSpecification;

	private List<Field> Fields;
	
	// Accessors
//	public List<TypeArgument> getTypeArguments() {
//		return TypeArguments;
//	}

	public List<Expression> getArguments() {
		return Arguments;
	}

//	public TypeSpecification getTypeSpecification() {
//		return TypeSpecification;
//	}

	/**
	 * @return List of fields for anonymous classes
	 */
	public List<Field> getFields() {
		return Fields;
	}
	
	public Type getSpecificationType() {
		return specType;
	}
	
	public int getDimensions() {
		return dimensions;
	}
	
	public List<ArrayInitializer> getArrayInitializer() {
		return ArrayInitializers;
	}

	protected void acceptChildren(Visitor visitor) {
		// Visiting non-text attributes
//		new CollectionVisitable(TypeArguments, false).accept(visitor);
		new CollectionVisitable(Arguments, false).accept(visitor);
		new CollectionVisitable(ArrayInitializers, false).accept(visitor);
		new CollectionVisitable(Fields, false).accept(visitor);
	}

	public MethodInfo findMethod(String name, String[] parameterTypes) {
		MethodInfo candidate = findTypeMethod(name, parameterTypes);
		
		if (candidate!=null) {
			return candidate; // Does Java look for more specific methods in enclosed scope?
		}

		Scope ec = findParent(Scope.class);
		return ec==null ? null : ec.findMethod(name, parameterTypes);
	}

	public TypeInfo findType(String name) {
		TypeInfo ret = findNestedType(name);
		if (ret!=null) {
			return ret;
		}

		Scope ec = findParent(Scope.class);
		return ec==null ? null : ec.findType(name);
	}

	public VariableInfo findVariable(String name) {
		VariableInfo ret = findTypeVariable(name);
		if (ret!=null) {
			return ret;
		}

		Scope ec = findParent(Scope.class);
		return ec==null ? null : ec.findVariable(name);
	}

	public boolean isKindOf(String superFcn) {
		TypeSpecInfo tsi = getTypeSpecInfo();
		if (tsi!=null) {
			TypeInfo ti = TypeSpecification.toTypeInfo(tsi, findParent(Scope.class));
			if (ti!=null) {
				return ti.isKindOf(superFcn);
			}
		}
		return JAVA_LANG_OBJECT.equals(superFcn);
	}

	public String getFcn() {
		if (!fcnCalculated) {
			super.getFcn();
			SourceUnit su = getSourceUnit();
			if (su instanceof JavaSourceUnit) {
				((JavaSourceUnit) su).registerLocalOrAnonymous(fcn, this);
			}		
		}
		
		return fcn;			
	}

	public TypeInfo findNestedType(String name) {
		for (Field field: getFields()) {
			if (field instanceof TypeDefinition) {
				final TypeDefinition td = (TypeDefinition) field;
				if (td.getName().equals(name)) {
					return td;
				}
			}
		}
		
		TypeSpecInfo tsi = getTypeSpecInfo();
		if (tsi==null || tsi.getTypeInfo()==null) {
			return null;
		}
		TypeInfo ti = TypeSpecification.toTypeInfo(tsi, findParent(Scope.class));
		return ti==null ? null : ti.findNestedType(name); 
	}

	public MethodInfo findTypeMethod(String name, String[] parameterTypes) {
		List<TypeSpecInfo> pTypes=new ArrayList<TypeSpecInfo>();
		boolean pTypesResolved = true;
		for (int i=0; parameterTypes!=null && i<parameterTypes.length; ++i) {
			if (parameterTypes[i]==null) {
				StringBuffer sig = new StringBuffer();
				for (String pt: parameterTypes) {
					sig.append(sig.length()==0 ? "(" : ", ");
					sig.append(pt);
				}
				sig.append(")");
				logger.warning("Could not resolve method "+name+sig.toString()+" argument "+i+" type ["+getLocation()+"]");
				pTypes.add(null);
				pTypesResolved = false;
			} else {
				int dimensions=0;
				String typeName = parameterTypes[i];
				while (typeName.endsWith(TypeSpecInfo.ARRAY_BRACKETS)) {
					++dimensions;
					typeName=typeName.substring(0, typeName.length()-TypeSpecInfo.ARRAY_BRACKETS.length());
				}
				final TypeInfo ptInfo = findType(typeName);
				if (ptInfo==null) {
					logger.warning("Type "+ptInfo+" not found");
					return null;
				}
				final int d = dimensions;
				pTypes.add(new TypeSpecInfo() {
	
					public int getDimensions() {
						return d;
					}
	
					public TypeInfo getTypeInfo() {
						return ptInfo;
					}
					
				});
			}
		}
	
		MethodInfo pnumMatch = null;
		MethodInfo candidate = null;
		for (Field field: getFields()) {
			if (field instanceof MethodDefinition) {
				MethodDefinition method = (MethodDefinition) field;
				if (!method.getName().equals(name)) {
					continue;
				}
				
				List<TypeSpecInfo> mTypes = method.getParameterTypes();				
				ParameterMatch matchResult = TypeDefinition.match(mTypes, pTypes);
				if (matchResult == ParameterMatch.FULL_MATCH) {
					return method;
				} else if (matchResult == ParameterMatch.MATCH) {
					candidate = TypeDefinition.selectMoreSpecific(candidate, method);
				} else if (!pTypesResolved && pnumMatch==null) {
					pnumMatch = method;
				}
			}
		}
		
		TypeSpecInfo tsi = getTypeSpecInfo();
		if (tsi==null || tsi.getTypeInfo()==null) {
			return candidate==null ? pnumMatch : candidate;
		}
		TypeInfo ti = TypeSpecification.toTypeInfo(tsi, findParent(Scope.class));
		if (ti==null) {
			return candidate==null ? pnumMatch : candidate;			
		}
		
		if (candidate==null && !pTypesResolved) {
			return pnumMatch;
		}
		
		return TypeDefinition.selectMoreSpecific(candidate, ti.findTypeMethod(name, parameterTypes)); 
	}

	public VariableInfo findTypeVariable(String name) {
		for (Field field: getFields()) {
			if (field instanceof VariableDefinition) {
				final VariableDefinition vd = (VariableDefinition) field;
				if (vd.getName().equals(name)) {
					return vd;
				}
			}
		}

		TypeSpecInfo tsi = getTypeSpecInfo();
		if (tsi==null || tsi.getTypeInfo()==null) {
			return null;
		}
		TypeInfo ti = TypeSpecification.toTypeInfo(tsi, findParent(Scope.class));
		return ti==null ? null : ti.findTypeVariable(name); 
	}


	public MethodInfo findConstructor(String[] argumentTypes) {
		// TODO ???
		return null;
	}
	
	public TypeInfo getDeclaringType() {
		return null;
	}

	public List<String> getModifiers() {
		return emptyList;
	}

	public String getName() {
		return null;
	}
	
	@Override
	public TypeSpecInfo getTypeSpecInfo() {
		final Info info;
		if (specType instanceof Identifier) {
			Scope es = findParent(Scope.class);
			info = ((Identifier) specType).getInfo(es);
		} else {
			 info = specType.getInfo();
		}
		
		return new TypeSpecInfo() {

			public int getDimensions() {
				return dimensions;
			}

			public TypeInfo getTypeInfo() {
				return info instanceof TypeInfo ? (TypeInfo) info : null;
			}
			
		};
	}

	public MethodInfo getInfo() {				
		Scope scope = findParent(Scope.class);
		if (scope == null) {
			return null;
		}
		
		List<Expression> args = getArguments();
		String[] argumentTypes = new String[args.size()];
		Iterator<Expression> it = args.iterator();
		for (int i=0; it.hasNext(); ++i) {
			argumentTypes[i] = TypeSpecification.toString(it.next().getTypeSpecInfo());
		}
		
		TypeInfo ti = TypeSpecification.toTypeInfo(getTypeSpecInfo(), scope);
		return ti==null ? null : ti.findConstructor(argumentTypes);		
	}
}
