package org.mesopotamia.lang.java;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

import org.mesopotamia.LanguageElement;
import org.mesopotamia.MesopotamiaException;
import org.mesopotamia.NodeData;
import org.mesopotamia.RepositoryLanguage;
import org.mesopotamia.Scan;
import org.mesopotamia.SimpleLanguageElement;
import org.mesopotamia.SourceUnit;
import org.mesopotamia.lang.java.ref.MethodInfo;
import org.mesopotamia.lang.java.ref.Scope;
import org.mesopotamia.lang.java.ref.SoftCachingTypeInfo;
import org.mesopotamia.lang.java.ref.TypeInfo;
import org.mesopotamia.lang.java.ref.TypeSpecInfo;
import org.mesopotamia.lang.java.ref.VariableInfo;
import org.w3c.dom.Element;

import biz.hammurapi.util.CollectionVisitable;
import biz.hammurapi.util.Visitor;

public class TypeDefinition extends Field implements TypeInfo, TypeDefinitionOrAnonymous {
	private static final Logger logger = Logger.getLogger(TypeDefinition.class.getName());
	
	public TypeDefinition(
			NodeData xData, 
			Class<?> context, 
			Scan scan,
			RepositoryLanguage language, 
			Object environment) throws MesopotamiaException {
		
		super(xData, context, scan, language, environment);

		// Select attributes
		Name = selectSingleElementText(Identifier.class, "IDENT|DOT");
		Fields = select(Field.class, "OBJBLOCK/*");

	}

	public void toDom(Element holder) {
		super.toDom(holder);

		// Serialize attributes
		setAttribute(holder, "Name", Name);
		setElement(holder, "Fields", Fields);
	}

	// Attributes
	private String Name;

	private List<Field> Fields;

	// Accessors
	public String getName() {
		return Name;
	}

	public List<Field> getFields() {
		return Fields;
	}

	protected void acceptChildren(Visitor visitor) {
		super.acceptChildren(visitor);
		// Visiting non-text attributes
		new CollectionVisitable(Fields, false).accept(visitor);
	}
	
	@SuppressWarnings("unchecked")
	public String getFcn() {
		if (!fcnCalculated) {
			LanguageElement parent = getParent();
			if (parent instanceof SimpleLanguageElement && "OBJBLOCK".equals(parent.getTokenName())) {
				parent = parent.getParent();
			}
			if (parent instanceof TypeDefinition) {
				fcn = ((TypeDefinition) parent).getFcn()+"."+getName();
			} else if (parent instanceof JavaLanguageElement) {
				fcn = ((JavaLanguageElement) parent).getFcn()+"."+getName();		
				SourceUnit su = getSourceUnit();
				if (su instanceof JavaSourceUnit) {
					((JavaSourceUnit) su).registerLocalOrAnonymous(fcn, this);
				}
			} else if (parent==null) {
				String packageName = ((JavaSourceUnit) getSourceUnit()).getPackageName();
				if (packageName.length()==0) {
					fcn = getName();
				} else {
					fcn = packageName+"."+getName();
				}
			} else {
				throw new IllegalStateException("Logical problem");
			}
			
			fcnCalculated = true;
		}
		return fcn;
	}
	
	public MethodInfo findMethod(String name, String[] parameterTypes) {
		MethodInfo candidate = findTypeMethod(name, parameterTypes);
		
		if (candidate!=null) {
			return candidate; // Does Java look for more specific methods in enclosed scope?
		}

		Scope ec = getEnclosingScope();
		return ec==null ? null : ec.findMethod(name, parameterTypes);
	}
	
	public static enum ParameterMatch {
		NO_MATCH,
		MATCH,
		FULL_MATCH;		
	}
	
	/**
	 * Matches formal parameter types with arguments for compatibility.
	 * @param parameters
	 * @param arguments
	 * @return Match result. Full match means that arguments are exactly of parmeter types, 
	 * match means that some of them are subtypes of parameter types.
	 */
	public static ParameterMatch match(List<TypeSpecInfo> parameters, List<TypeSpecInfo> arguments) {
		if (parameters.size()!=arguments.size()) {
			return ParameterMatch.NO_MATCH;
		}
		
		boolean fullMatch = true;
		Iterator<TypeSpecInfo> ppit = parameters.iterator();
		Iterator<TypeSpecInfo> apit = arguments.iterator();
		while (apit.hasNext() && ppit.hasNext()) {
			TypeSpecInfo argument = apit.next();
			if (argument == null) {
				return ParameterMatch.NO_MATCH;
			}
			TypeSpecInfo parameter = ppit.next();
			if (parameter == null) {
				return ParameterMatch.NO_MATCH;
			}
			if (argument!=Null.NULL_TYPE_SPEC_INFO && argument.getDimensions()!=parameter.getDimensions()) {
				if (parameter.getDimensions()==0 
						&& parameter.getTypeInfo()!=null 
						&& JAVA_LANG_OBJECT.equals(parameter.getTypeInfo().getFcn())) {
					fullMatch = false;
					continue;
				}
				return ParameterMatch.NO_MATCH;
			}
			
			TypeInfo ati = argument.getTypeInfo();
			if (ati==null) {
				return ParameterMatch.NO_MATCH;				
			}
			TypeInfo pti = parameter.getTypeInfo();
			if (pti == null) {
				return ParameterMatch.NO_MATCH;				
			}				
			if (!ati.isKindOf(pti.getFcn())) {
				return ParameterMatch.NO_MATCH;
			}
			if (!pti.getFcn().equals(ati.getFcn())) {
				fullMatch = false;
			}
		}
		return fullMatch ? ParameterMatch.FULL_MATCH : ParameterMatch.MATCH;		
	}
	
	/**
	 * @param m0
	 * @param m1
	 * @return m0 or m1 if m1 is more specific than m0. This method doesn't handle ambiguity.
	 */
	public static MethodInfo selectMoreSpecific(MethodInfo m0, MethodInfo m1) {
		if (m0==null) {
			return m1;
		} 
		if (m1==null) {
			return m0;
		}
		
		ParameterMatch cmr = match(m0.getParameterTypes(), m1.getParameterTypes());
		return cmr==ParameterMatch.MATCH ? m1 : m0;
	}
	
	protected Scope getEnclosingScope() {
		if (getParent()==null) {
			SourceUnit su = getSourceUnit();
			return su instanceof JavaSourceUnit ? (JavaSourceUnit) su : null;
		} 

		return findParent(Scope.class);
	}

	public TypeInfo findType(String name) {
		TypeInfo ret = findNestedType(name);
		if (ret!=null) {
			return ret;
		}

		Scope ec = getEnclosingScope();
		return ec==null ? null : ec.findType(name);
	}

	public VariableInfo findVariable(String name) {
		VariableInfo ret = findTypeVariable(name);
		if (ret!=null) {
			return ret;
		}

		Scope ec = getEnclosingScope();
		return ec==null ? null : ec.findVariable(name);
	}

	protected boolean isKindOfInternal(String superFcn) {
		return JAVA_LANG_OBJECT.equals(superFcn) || superFcn.equals(getFcn());
	}

	protected TypeInfo findNestedTypeInternal(String name) {
		for (Field field: getFields()) {
			if (field instanceof TypeDefinition) {
				final TypeDefinition td = (TypeDefinition) field;
				if (td.getName().equals(name) || td.getFcn().equals(name)) {
					return td;
				}
			}
		}
		
		return null; // Subclasses traverse supers
	}

	protected MethodInfo findTypeMethodInternal(String name, String[] argumentTypes) {
		List<TypeSpecInfo> aTypes=new ArrayList<TypeSpecInfo>();
		boolean pTypesResolved = true;
		for (int i=0; argumentTypes!=null && i<argumentTypes.length; ++i) {
			if (argumentTypes[i]==null) {
				StringBuffer sig = new StringBuffer();
				for (String pt: argumentTypes) {
					sig.append(sig.length()==0 ? "(" : ", ");
					sig.append(pt);
				}
				sig.append(")");
				logger.warning("Could not resolve method "+name+sig.toString()+" argument "+i+" type ["+getFcn()+", "+getLocation()+"]");
				aTypes.add(null);
				pTypesResolved = false;
			} else {
				int dimensions=0;
				String typeName = argumentTypes[i];
				while (typeName.endsWith(TypeSpecInfo.ARRAY_BRACKETS)) {
					++dimensions;
					typeName=typeName.substring(0, typeName.length()-TypeSpecInfo.ARRAY_BRACKETS.length());
				}
				final TypeInfo ptInfo = findType(typeName);
				if (ptInfo==null) {
					logger.warning("Type "+ptInfo+" not found");
					return null;
				}
				final int d = dimensions;
				aTypes.add(new TypeSpecInfo() {
	
					public int getDimensions() {
						return d;
					}
	
					public TypeInfo getTypeInfo() {
						return ptInfo;
					}
					
				});
			}
		}
	
		MethodInfo pnumMatch = null;
		MethodInfo candidate = null;
		for (Field field: getFields()) {
			if (field instanceof MethodDefinition) {
				MethodDefinition method = (MethodDefinition) field;
				if (!method.getName().equals(name)) {
					continue;
				}
				
				List<TypeSpecInfo> mTypes = method.getParameterTypes();				
				ParameterMatch matchResult = match(mTypes, aTypes);
				if (matchResult == ParameterMatch.FULL_MATCH) {
					return method;
				} else if (matchResult == ParameterMatch.MATCH) {
					candidate = selectMoreSpecific(candidate, method);
				} else if (!pTypesResolved && pnumMatch==null) {
					pnumMatch = method;
				}
			}
		}
		
		return candidate==null ? pnumMatch : candidate;
	}

	protected VariableInfo findTypeVariableInternal(String name) {
		for (Field field: getFields()) {
			if (field instanceof VariableDefinition) {
				final VariableDefinition vd = (VariableDefinition) field;
				if (vd.getName().equals(name)) {
					return vd;
				}
			}
		}

		return null; // Subclasses traverse supers
	}

	public TypeInfo getDeclaringType() {
		LanguageElement parent = getParent();
		return parent instanceof TypeDefinition ? (TypeDefinition) parent : null;
	}
	
	private TypeInfo cachingTypeInfo = new SoftCachingTypeInfo() {

		@Override
		protected MethodInfo findConstructorInternal(String[] argumentTypes) {
			boolean pTypesResolved = true;
			List<TypeSpecInfo> aTypes=new ArrayList<TypeSpecInfo>();
			for (int i=0; argumentTypes!=null && i<argumentTypes.length; ++i) {
				if (argumentTypes[i]==null) {
					StringBuffer sig = new StringBuffer();
					for (String pt: argumentTypes) {
						sig.append(sig.length()==0 ? "(" : ", ");
						sig.append(pt);
					}
					sig.append(")");
					logger.warning("Could not resolve constructor "+getName()+sig.toString()+" argument "+i+" type ["+getFcn()+", "+getLocation()+"]");
					aTypes.add(null);
					pTypesResolved = false;
				} else {
					int dimensions=0;
					String typeName = argumentTypes[i];
					while (typeName.endsWith(TypeSpecInfo.ARRAY_BRACKETS)) {
						++dimensions;
						typeName=typeName.substring(0, typeName.length()-TypeSpecInfo.ARRAY_BRACKETS.length());
					}
					final TypeInfo ptInfo = findType(typeName);
					if (ptInfo==null) {
						logger.warning("Type "+ptInfo+" not found");
						return null;
					}
					final int d = dimensions;
					aTypes.add(new TypeSpecInfo() {
		
						public int getDimensions() {
							return d;
						}
		
						public TypeInfo getTypeInfo() {
							return ptInfo;
						}
						
					});
				}
			}
		
			MethodInfo pnumMatch = null;
			MethodInfo candidate = null;
			for (Field field: getFields()) {
				if (field instanceof ConstructorDefinition) {
					ConstructorDefinition constructor = (ConstructorDefinition) field;
					
					List<TypeSpecInfo> mTypes = constructor.getParameterTypes();				
					ParameterMatch matchResult = match(mTypes, aTypes);
					if (matchResult == ParameterMatch.FULL_MATCH) {
						return constructor;
					} else if (matchResult == ParameterMatch.MATCH) {
						candidate = selectMoreSpecific(candidate, constructor);
					} else if (!pTypesResolved && pnumMatch==null) {
						pnumMatch = constructor;
					}
				}
			}
			
			return candidate == null ? pnumMatch : candidate;
		}

		@Override
		protected MethodInfo findMethodInternal(String name, String[] parameterTypes) {
			return TypeDefinition.this.findTypeMethodInternal(name, parameterTypes);
		}

		@Override
		protected TypeInfo findNestedTypeInternal(String name) {
			return TypeDefinition.this.findNestedTypeInternal(name);
		}

		@Override
		protected VariableInfo findTypeVariableInternal(String name) {
			return TypeDefinition.this.findTypeVariableInternal(name);
		}

		@Override
		protected TypeInfo getDeclaringTypeInternal() {
			throw new UnsupportedOperationException();
		}

		@Override
		protected String getFcnInternal() {
			throw new UnsupportedOperationException();
		}

		@Override
		protected List<String> getModifiersInternal() {
			throw new UnsupportedOperationException();
		}

		@Override
		protected boolean isKindOfInternal(String superFcn) {
			return TypeDefinition.this.isKindOfInternal(superFcn);
		}

		public String getName() {
			return TypeDefinition.this.getName();
		}
		
	};

	public MethodInfo findConstructor(String[] argumentTypes) {
		return cachingTypeInfo.findConstructor(argumentTypes);
	}

	public TypeInfo findNestedType(String name) {
		return cachingTypeInfo.findNestedType(name);
	}

	public MethodInfo findTypeMethod(String name, String[] argumentTypes) {
		return cachingTypeInfo.findTypeMethod(name, argumentTypes);
	}

	public VariableInfo findTypeVariable(String name) {
		return cachingTypeInfo.findTypeVariable(name);
	}

	public boolean isKindOf(String superFcn) {
		return cachingTypeInfo.isKindOf(superFcn);
	}

}
