package org.mesopotamia.lang.java.ref;

import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Scope which caches resolved types, variables and methods
 * using soft references.
 * @author Pavel
 */
public abstract class SoftCachingScope implements Scope {
	
	private static final MethodInfo NO_METHOD_INFO = new MethodInfo() {

		public List<TypeInfo> getExceptionTypes() {
			throw new UnsupportedOperationException();
		}

		public List<TypeSpecInfo> getParameterTypes() {
			throw new UnsupportedOperationException();
		}

		public TypeSpecInfo getReturnType() {
			throw new UnsupportedOperationException();
		}

		public TypeInfo getDeclaringType() {
			throw new UnsupportedOperationException();
		}

		public String getFcn() {
			throw new UnsupportedOperationException();
		}

		public List<String> getModifiers() {
			throw new UnsupportedOperationException();
		}

		public String getName() {
			throw new UnsupportedOperationException();
		}
		
	};

	private Map<List<String>, SoftReference<MethodInfo>> methodInfoCache = new ConcurrentHashMap<List<String>, SoftReference<MethodInfo>>();
	
	public MethodInfo findMethod(String name, String[] parameterTypes) {
		List<String> key = new ArrayList<String>();
		key.add(name);
		if (parameterTypes!=null) {
			for (String pt: parameterTypes) {
				key.add(pt);
			}
		}
		
		SoftReference<MethodInfo> ref = methodInfoCache.get(key);
		MethodInfo ret = ref==null ? null : ref.get();
		if (ret == null) {
			ret = findMethodInternal(name, parameterTypes);
			if (ret == null) {
				ret = NO_METHOD_INFO;
			}
			methodInfoCache.put(key, new SoftReference<MethodInfo>(ret));
		}
		return ret==NO_METHOD_INFO ? null : ret;
	}
	
	protected abstract MethodInfo findMethodInternal(String name, String[] parameterTypes);

	private static final TypeInfo NO_TYPE_INFO = new TypeInfo() {

		public MethodInfo findConstructor(String[] argumentTypes) {
			throw new UnsupportedOperationException();
		}

		public TypeInfo findNestedType(String name) {
			throw new UnsupportedOperationException();
		}

		public MethodInfo findTypeMethod(String name, String[] argumentTypes) {
			throw new UnsupportedOperationException();
		}

		public VariableInfo findTypeVariable(String name) {
			throw new UnsupportedOperationException();
		}

		public boolean isKindOf(String superFcn) {
			throw new UnsupportedOperationException();
		}

		public TypeInfo getDeclaringType() {
			throw new UnsupportedOperationException();
		}

		public String getFcn() {
			throw new UnsupportedOperationException();
		}

		public List<String> getModifiers() {
			throw new UnsupportedOperationException();
		}

		public String getName() {
			throw new UnsupportedOperationException();
		}
		
	};
	
	private Map<String, SoftReference<TypeInfo>> typeInfoCache = new ConcurrentHashMap<String, SoftReference<TypeInfo>>();
	
	public TypeInfo findType(String name) {
		SoftReference<TypeInfo> ref = typeInfoCache.get(name);
		TypeInfo ret = ref==null ? null : ref.get();
		if (ret == null) {
			ret = findTypeInternal(name);
			if (ret == null) {
				ret = NO_TYPE_INFO;
			}
			typeInfoCache.put(name, new SoftReference<TypeInfo>(ret));
		}
		return ret==NO_TYPE_INFO ? null : ret;
	}	

	protected abstract TypeInfo findTypeInternal(String name);

	private static final VariableInfo NO_VARIABLE_INFO = new VariableInfo() {

		public TypeSpecInfo getTypeSpecification() {
			throw new UnsupportedOperationException();
		}

		public TypeInfo getDeclaringType() {
			throw new UnsupportedOperationException();
		}

		public String getFcn() {
			throw new UnsupportedOperationException();
		}

		public List<String> getModifiers() {
			throw new UnsupportedOperationException();
		}

		public String getName() {
			throw new UnsupportedOperationException();
		}
		
	};
	
	private Map<String, SoftReference<VariableInfo>> variableInfoCache = new ConcurrentHashMap<String, SoftReference<VariableInfo>>();
	
	public VariableInfo findVariable(String name) {
		SoftReference<VariableInfo> ref = variableInfoCache.get(name);
		VariableInfo ret = ref==null ? null : ref.get();
		if (ret == null) {
			ret = findVariableInternal(name);
			if (ret == null) {
				ret = NO_VARIABLE_INFO;
			}
			variableInfoCache.put(name, new SoftReference<VariableInfo>(ret));
		}
		return ret==NO_VARIABLE_INFO ? null : ret;
	}

	protected abstract VariableInfo findVariableInternal(String name);
	
}
