package org.mesopotamia.lang.java.ref;

import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Type info which caches resolved types, methods, and variables
 * in soft references and other values in hard references.
 * @author Pavel
 *
 */
public abstract class SoftCachingTypeInfo implements TypeInfo {
		
	private static final MethodInfo NO_METHOD_INFO = new MethodInfo() {

		public List<TypeInfo> getExceptionTypes() {
			throw new UnsupportedOperationException();
		}

		public List<TypeSpecInfo> getParameterTypes() {
			throw new UnsupportedOperationException();
		}

		public TypeSpecInfo getReturnType() {
			throw new UnsupportedOperationException();
		}

		public TypeInfo getDeclaringType() {
			throw new UnsupportedOperationException();
		}

		public String getFcn() {
			throw new UnsupportedOperationException();
		}

		public List<String> getModifiers() {
			throw new UnsupportedOperationException();
		}

		public String getName() {
			throw new UnsupportedOperationException();
		}
		
	};

	private Map<List<String>, SoftReference<MethodInfo>> methodInfoCache = new ConcurrentHashMap<List<String>, SoftReference<MethodInfo>>();
	
	public MethodInfo findTypeMethod(String name, String[] parameterTypes) {
		List<String> key = new ArrayList<String>();
		key.add(name);
		if (parameterTypes!=null) {
			for (String pt: parameterTypes) {
				key.add(pt);
			}
		}
		
		SoftReference<MethodInfo> ref = methodInfoCache.get(key);
		MethodInfo ret = ref==null ? null : ref.get();
		if (ret == null) {
			ret = findMethodInternal(name, parameterTypes);
			if (ret == null) {
				ret = NO_METHOD_INFO;
			}
			methodInfoCache.put(key, new SoftReference<MethodInfo>(ret));
		}
		return ret==NO_METHOD_INFO ? null : ret;
	}
	
	private Map<List<String>, SoftReference<MethodInfo>> constructorInfoCache = new ConcurrentHashMap<List<String>, SoftReference<MethodInfo>>();
	
	public MethodInfo findConstructor(String[] parameterTypes) {
		List<String> key = new ArrayList<String>();
		if (parameterTypes!=null) {
			for (String pt: parameterTypes) {
				key.add(pt);
			}
		}
		
		SoftReference<MethodInfo> ref = constructorInfoCache.get(key);
		MethodInfo ret = ref==null ? null : ref.get();
		if (ret == null) {
			ret = findConstructorInternal(parameterTypes);
			if (ret == null) {
				ret = NO_METHOD_INFO;
			}
			constructorInfoCache.put(key, new SoftReference<MethodInfo>(ret));
		}
		return ret==NO_METHOD_INFO ? null : ret;
	}
	
	protected abstract MethodInfo findConstructorInternal(String[] parameterTypes);
	protected abstract MethodInfo findMethodInternal(String name, String[] parameterTypes);

	private static final TypeInfo NO_TYPE_INFO = new TypeInfo() {

		public MethodInfo findConstructor(String[] argumentTypes) {
			throw new UnsupportedOperationException();
		}

		public TypeInfo findNestedType(String name) {
			throw new UnsupportedOperationException();
		}

		public MethodInfo findTypeMethod(String name, String[] argumentTypes) {
			throw new UnsupportedOperationException();
		}

		public VariableInfo findTypeVariable(String name) {
			throw new UnsupportedOperationException();
		}

		public boolean isKindOf(String superFcn) {
			throw new UnsupportedOperationException();
		}

		public TypeInfo getDeclaringType() {
			throw new UnsupportedOperationException();
		}

		public String getFcn() {
			throw new UnsupportedOperationException();
		}

		public List<String> getModifiers() {
			throw new UnsupportedOperationException();
		}

		public String getName() {
			throw new UnsupportedOperationException();
		}
		
	};
	
	private Map<String, SoftReference<TypeInfo>> typeInfoCache = new ConcurrentHashMap<String, SoftReference<TypeInfo>>();
	
	public TypeInfo findNestedType(String name) {
		SoftReference<TypeInfo> ref = typeInfoCache.get(name);
		TypeInfo ret = ref==null ? null : ref.get();
		if (ret == null) {
			ret = findNestedTypeInternal(name);
			if (ret == null) {
				ret = NO_TYPE_INFO;
			}
			typeInfoCache.put(name, new SoftReference<TypeInfo>(ret));
		}
		return ret==NO_TYPE_INFO ? null : ret;
	}	

	protected abstract TypeInfo findNestedTypeInternal(String name);

	private static final VariableInfo NO_VARIABLE_INFO = new VariableInfo() {

		public TypeSpecInfo getTypeSpecification() {
			throw new UnsupportedOperationException();
		}

		public TypeInfo getDeclaringType() {
			throw new UnsupportedOperationException();
		}

		public String getFcn() {
			throw new UnsupportedOperationException();
		}

		public List<String> getModifiers() {
			throw new UnsupportedOperationException();
		}

		public String getName() {
			throw new UnsupportedOperationException();
		}
		
	};
	
	private Map<String, SoftReference<VariableInfo>> variableInfoCache = new ConcurrentHashMap<String, SoftReference<VariableInfo>>();
	
	public VariableInfo findTypeVariable(String name) {
		SoftReference<VariableInfo> ref = variableInfoCache.get(name);
		VariableInfo ret = ref==null ? null : ref.get();
		if (ret == null) {
			ret = findTypeVariableInternal(name);
			if (ret == null) {
				ret = NO_VARIABLE_INFO;
			}
			variableInfoCache.put(name, new SoftReference<VariableInfo>(ret));
		}
		return ret==NO_VARIABLE_INFO ? null : ret;
	}

	protected abstract VariableInfo findTypeVariableInternal(String name);
	
	private Map<String, Boolean> isKindOfMap = new HashMap<String, Boolean>();

	public boolean isKindOf(String superFcn) {
		Boolean ret = isKindOfMap.get(superFcn);
		if (ret==null) {
			ret = isKindOfInternal(superFcn) ? Boolean.TRUE : Boolean.FALSE;
			isKindOfMap.put(superFcn, ret);
		}
		return ret.booleanValue();
	}

	protected abstract boolean isKindOfInternal(String superFcn);

	private TypeInfo declaringType;
	private boolean isDeclaringTypeRetrieved;
	protected abstract TypeInfo getDeclaringTypeInternal();	
	
	public TypeInfo getDeclaringType() {
		if (!isDeclaringTypeRetrieved) {
			declaringType = getDeclaringTypeInternal();
			isDeclaringTypeRetrieved = true;
		}
		return declaringType;
	}
	
	private String fcn;
	private boolean isFcnRetrieved;
	protected abstract String getFcnInternal();

	public String getFcn() {
		if (!isFcnRetrieved) {
			fcn = getFcnInternal();
			isFcnRetrieved = true;
		}
		return fcn;
	}

	private List<String> modifiers;
	private boolean areModifiersRetrieved;
	protected abstract List<String> getModifiersInternal();
	
	public List<String> getModifiers() {
		if (!areModifiersRetrieved) {
			modifiers = getModifiersInternal();
			areModifiersRetrieved = true;
		}
		return modifiers;
	}

}
