package org.mesopotamia.lang.java.ref.bcel;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.Type;
import org.mesopotamia.lang.java.BuiltInType;
import org.mesopotamia.lang.java.Null;
import org.mesopotamia.lang.java.TypeDefinition;
import org.mesopotamia.lang.java.TypeSpecification;
import org.mesopotamia.lang.java.TypeDefinition.ParameterMatch;
import org.mesopotamia.lang.java.ref.MethodInfo;
import org.mesopotamia.lang.java.ref.Scope;
import org.mesopotamia.lang.java.ref.TypeInfo;
import org.mesopotamia.lang.java.ref.TypeSpecInfo;
import org.mesopotamia.lang.java.ref.VariableInfo;

public class JavaClassTypeInfo implements TypeInfo {
	
	private static final String JAVA_LANG_OBJECT = "java.lang.Object";

	private static final Logger logger = Logger.getLogger(JavaClassTypeInfo.class.getName());
	
	private JavaClass javaClass;

	private Scope scope;

	private JavaClassTypeInfoRepository repository;

	JavaClassTypeInfo(JavaClass javaClass, Scope scope, JavaClassTypeInfoRepository repository) {
		this.javaClass = javaClass;
		this.scope = scope;
		this.repository = repository;
	}

	public TypeInfo findNestedType(String name) {
		String nName = javaClass.getClassName()+"$"+name;
		TypeInfo ret = repository.getTypeInfo(nName, scope);
		if (ret!=null) {
			return ret;
		}
		
        if (JAVA_LANG_OBJECT.equals(javaClass.getClassName())) {
            return null;
        }		
		
		String superClassName = javaClass.getSuperclassName();
		TypeInfo superInfo = repository.getTypeInfo(superClassName, scope);
		if (superInfo!=null) {
			TypeInfo sret = superInfo.findNestedType(name);
			if (sret!=null) {
				return sret;
			}
		}
		
		for (String interfaceName: javaClass.getInterfaceNames()) {
			TypeInfo iInfo = repository.getTypeInfo(interfaceName, scope);
			if (iInfo!=null) {
				TypeInfo iret = iInfo.findNestedType(name);
				if (iret!=null) {
					return iret;
				}
			}
		}										
		
		return null;
	}
	
	public MethodInfo findTypeMethod(String name, String[] parameterTypes) {
		List<TypeSpecInfo> pTypes=new ArrayList<TypeSpecInfo>();
		boolean pTypesResolved = true;
		for (int i=0; parameterTypes!=null && i<parameterTypes.length; ++i) {
			if (parameterTypes[i]==null) {
				StringBuffer sig = new StringBuffer();
				for (String pt: parameterTypes) {
					sig.append(sig.length()==0 ? "(" : ", ");
					sig.append(pt);
				}
				sig.append(")");
				logger.warning("Could not resolve method "+name+sig.toString()+" argument "+i+" type ["+javaClass.getClassName()+"]");
				pTypes.add(null);
				pTypesResolved = false;
			} else {
				pTypes.add(string2TypeSpecInfo(parameterTypes[i], scope, repository));
			}
		}
		
		MethodInfo candidate = null;
		MethodInfo pnumMatch = null;
		Method[] ma = javaClass.getMethods();
		for (int i=0; i<ma.length; ++i) {
			if (!ma[i].getName().equals(name)) {
				continue;
			}
			
			Type[] at = ma[i].getArgumentTypes();
			if ((parameterTypes==null || parameterTypes.length==0) && (at==null || at.length==0)) {
				return new JavaClassMethodInfo(ma[i], scope, repository, this);				
			}
			
			int pl = parameterTypes==null ? 0 : parameterTypes.length;
			int al = at==null ? 0 : at.length;
			if (pl!=al) {
				continue;
			}			
			
			List<TypeSpecInfo> mTypes = new ArrayList<TypeSpecInfo>();
			for (int j=0; j<at.length; ++j) {
				mTypes.add(type2TypeSpecInfo(at[j], scope, repository));
			}
			ParameterMatch matchResult = TypeDefinition.match(mTypes, pTypes);
			if (matchResult == ParameterMatch.FULL_MATCH) {
				return new JavaClassMethodInfo(ma[i], scope, repository, this);
			} else if (matchResult == ParameterMatch.MATCH) {
				candidate = TypeDefinition.selectMoreSpecific(candidate, new JavaClassMethodInfo(ma[i], scope, repository, this));
			} else if (!pTypesResolved && pnumMatch==null) {
				pnumMatch = new JavaClassMethodInfo(ma[i], scope, repository, this);
			}
		}
		
		if (candidate==null && !pTypesResolved && pnumMatch!=null) {
			return pnumMatch;
		}
		
        if (JAVA_LANG_OBJECT.equals(javaClass.getClassName())) {
            return candidate;
        }
        
		String superClassName = javaClass.getSuperclassName();
		TypeInfo superInfo = repository.getTypeInfo(superClassName, scope);
		if (superInfo!=null) {
			candidate = TypeDefinition.selectMoreSpecific(candidate, superInfo.findTypeMethod(name, parameterTypes));
		}
		for (String interfaceName: javaClass.getInterfaceNames()) {
			TypeInfo iInfo = repository.getTypeInfo(interfaceName, scope);
			if (iInfo!=null) {
				candidate = TypeDefinition.selectMoreSpecific(candidate, iInfo.findTypeMethod(name, parameterTypes));
			}
		}
		
		return candidate;
	}

	public MethodInfo findConstructor(String[] parameterTypes) {
		List<TypeSpecInfo> pTypes=new ArrayList<TypeSpecInfo>();
		boolean pTypesResolved = true;
		for (int i=0; parameterTypes!=null && i<parameterTypes.length; ++i) {
			if (parameterTypes[i]==null) {
				StringBuffer sig = new StringBuffer();
				for (String pt: parameterTypes) {
					sig.append(sig.length()==0 ? "(" : ", ");
					sig.append(pt);
				}
				sig.append(")");
				logger.warning("Could not resolve constructor <init>"+sig.toString()+" argument "+i+" type ["+javaClass.getClassName()+"]");
				pTypes.add(null);
				pTypesResolved = false;
			} else {
				pTypes.add(string2TypeSpecInfo(parameterTypes[i], scope, repository));
			}
		}
		
		MethodInfo candidate = null;
		MethodInfo pnumMatch = null;
		Method[] ma = javaClass.getMethods();
		for (int i=0; i<ma.length; ++i) {
			if (!ma[i].getName().equals(MethodInfo.CONSTRUCTOR_NAME)) {
				continue;
			}
			
			Type[] at = ma[i].getArgumentTypes();
			List<TypeSpecInfo> mTypes = new ArrayList<TypeSpecInfo>();
			for (int j=0; j<at.length; ++j) {
				mTypes.add(type2TypeSpecInfo(at[j], scope, repository));
			}
			ParameterMatch matchResult = TypeDefinition.match(mTypes, pTypes);
			if (matchResult == ParameterMatch.FULL_MATCH) {
				return new JavaClassMethodInfo(ma[i], scope, repository, this);
			} else if (matchResult == ParameterMatch.MATCH) {
				candidate = TypeDefinition.selectMoreSpecific(candidate, new JavaClassMethodInfo(ma[i], scope, repository, this));
			} else if (!pTypesResolved && pnumMatch==null) {
				pnumMatch = new JavaClassMethodInfo(ma[i], scope, repository, this);
			}
		}
		
		return candidate==null ? pnumMatch : candidate;
	}

	public VariableInfo findTypeVariable(String name) {
		Field[] fields = javaClass.getFields();
		for (int i=0; i<fields.length; ++i) {
			if (fields[i].getName().equals(name)) {
				return new JavaClassVariableInfo(fields[i], scope, repository, this);
			}
		}
		
        if (JAVA_LANG_OBJECT.equals(javaClass.getClassName())) {
            return null;
        }
        
		String superClassName = javaClass.getSuperclassName();
		TypeInfo superInfo = repository.getTypeInfo(superClassName, scope);
		if (superInfo!=null) {
			VariableInfo ret = superInfo.findTypeVariable(name);
			if (ret!=null) {
				return ret;
			}
		}
		for (String interfaceName: javaClass.getInterfaceNames()) {
			TypeInfo iInfo = repository.getTypeInfo(interfaceName, scope);
			if (iInfo!=null) {
				VariableInfo ret = iInfo.findTypeVariable(name);
				if (ret!=null) {
					return ret;
				}
			}
		}

		return null;
	}

	public boolean isKindOf(String superFcn) {
		if (javaClass.getClassName().replace('$', '.').equals(superFcn)) { // Decide on $ or . in class names.
			return true;
		}
		
        if (JAVA_LANG_OBJECT.equals(javaClass.getClassName())) {
            return false;
        }
        
		String superClassName = javaClass.getSuperclassName();
		TypeInfo superInfo = repository.getTypeInfo(superClassName, scope);
		if (superInfo!=null && superInfo.isKindOf(superFcn)) {
			return true;
		}
		for (String interfaceName: javaClass.getInterfaceNames()) {
			TypeInfo iInfo = repository.getTypeInfo(interfaceName, scope);
			if (iInfo!=null && iInfo.isKindOf(superFcn)) {
				return true;
			}
		}

		return false;
	}

	public TypeInfo getDeclaringType() {
		String name = javaClass.getClassName();
		int idx = name.indexOf("$");
		if (idx==-1) {
			return null;
		}
		return repository.getTypeInfo(name.substring(0, idx), scope);
	}

	public String getFcn() {
		return javaClass.getClassName().replace('$', '.');
	}

	public List<String> getModifiers() {
		return modifiersList(javaClass.getModifiers());
	}
		
	public String getName() {
		String ret = javaClass.getClassName();
		int idx = ret.lastIndexOf('.');
		return idx == -1 ? ret : ret.substring(idx+1);
	}
	
	public static List<String> modifiersList(int mod) {
		List<String> ret = new ArrayList<String>();

		if ((mod & Modifier.PUBLIC) != 0) {
			ret.add("public ");
		}
		if ((mod & Modifier.PROTECTED) != 0) {
			ret.add("protected ");
		}
		if ((mod & Modifier.PRIVATE) != 0) {
			ret.add("private ");
		}

		/* Canonical order */
		if ((mod & Modifier.ABSTRACT) != 0) {
			ret.add("abstract ");
		}
		if ((mod & Modifier.STATIC) != 0) {
			ret.add("static ");
		}
		if ((mod & Modifier.FINAL) != 0) {
			ret.add("final ");
		}
		if ((mod & Modifier.TRANSIENT) != 0) {
			ret.add("transient ");
		}
		if ((mod & Modifier.VOLATILE) != 0) {
			ret.add("volatile ");
		}
		if ((mod & Modifier.SYNCHRONIZED) != 0) {
			ret.add("synchronized ");
		}
		if ((mod & Modifier.NATIVE) != 0) {
			ret.add("native ");
		}
		if ((mod & Modifier.STRICT) != 0) {
			ret.add("strictfp ");
		}
		if ((mod & Modifier.INTERFACE) != 0) {
			ret.add("interface ");
		}

		return ret;		
	}

	/**
	 * 
	 * @param type
	 * @param repository
	 */
	public static TypeSpecInfo type2TypeSpecInfo(Type type, Scope scope, JavaClassTypeInfoRepository repository) {
		return string2TypeSpecInfo(type.toString(), scope, repository);
	}
		
	public static TypeSpecInfo string2TypeSpecInfo(String typeName, Scope scope, JavaClassTypeInfoRepository repository) {	
		if (typeName==null) {
			return null;
		}
		
		if (Null.NULL_FCN.equals(typeName)) {
			return Null.NULL_TYPE_SPEC_INFO;
		}
		
		int dimensions=0;
		while (typeName.endsWith(TypeSpecInfo.ARRAY_BRACKETS)) {
			++dimensions;
			typeName=typeName.substring(0, typeName.length()-TypeSpecInfo.ARRAY_BRACKETS.length());
		}
		
		TypeInfo ti = BuiltInType.getBuiltInType(typeName);
		if (ti==null && scope!=null) {
			ti = scope.findType(typeName);
		}
		
		if (ti==null) {
			ti = repository.getTypeInfo(typeName, scope);
		}
		
		if (ti==null) {
			return null;
		}
		
		final int d = dimensions;
		final TypeInfo fti = ti;
		return new TypeSpecInfo() {

			public int getDimensions() {
				return d;
			}

			public TypeInfo getTypeInfo() {
				return fti;
			}
			
			@Override
			public String toString() {
				return TypeSpecification.toString(this);
			}
			
		};
		
	}
	
	@Override
	public String toString() {
		return javaClass.getClassName()+" ["+getClass().getName()+"]";
	}
}
