/*
  * mesopotamia-java @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia.lang.java.v13;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Reader;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.mesopotamia.LoaderBase;
import org.mesopotamia.LoaderEntry;
import org.mesopotamia.MesopotamiaException;
import org.mesopotamia.MesopotamiaToken;
import org.mesopotamia.RepositoryFactory;
import org.mesopotamia.RepositoryLanguage;
import org.mesopotamia.Source;
import org.mesopotamia.SourceLoader;
import org.mesopotamia.lang.java.CheatingFilterReader;
import org.mesopotamia.sql.LoadLevel;
import org.mesopotamia.sql.LoadLevelImpl;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import antlr.CharStreamException;
import antlr.Token;
import antlr.TokenStreamException;
import biz.hammurapi.convert.ConvertingService;
import biz.hammurapi.xml.dom.AbstractDomObject;


public class TokenLoader extends LoaderBase implements SourceLoader {
	private static final Logger logger = Logger.getLogger(TokenLoader.class.getName());

	public TokenLoader(RepositoryLanguage repoLanguage, LoaderEntry xData) {
		super(repoLanguage, xData);
	}

	public boolean load(int scanId, int sourceUnitId, Source source, Object environment) {
		RepositoryFactory factory=repoLanguage.getFactory();
		try {
			LoadLevelImpl lli = new LoadLevelImpl(true);
			lli.setLevelId(data.getId());
			lli.setSourceUnitId(sourceUnitId);
			try {
				Object rSource = source.get();
				Reader r = (Reader) ConvertingService.convert(rSource, Reader.class);
				try {
					// CheatingFilterReader is needed because of a bug in Java
					// gramma - it stucks if last token in the file
					// is SL_COMMENT

					JavaLexer lexer = new JavaLexer(new CheatingFilterReader(r)) {
						/**
						 * Replacement for generated mIDENT to recognize international characters.
						 */
						public Token nextToken() throws TokenStreamException {
					        try {
						        if (Character.isJavaIdentifierStart(LA(1))) {
						            resetText();
						            int _begin = text.length();
							        consume();
							        
							        while (Character.isJavaIdentifierPart(LA(1))) {
							            consume();
							        }
							
						            Token ret = makeToken(testLiteralsTable(IDENT));
						            ret.setText(new String(text.getBuffer(), _begin, text.length() - _begin));
						            
						            return ret;
								}
						        
								return super.nextToken();
					        } catch (CharStreamException e) {
					            throw new TokenStreamException(e.toString());
					        }
						}
					};
					
				    Document doc=DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
				    Element rootElement = AbstractDomObject.addElement(doc, "tokens");
					
					Token token;
					for (int i = 0; (token = lexer.nextToken()).getType() != JavaTokenTypes.EOF; i++) {
						Element te = doc.createElement("token");
						rootElement.appendChild(te);
						te.setAttribute("col", String.valueOf(token.getColumn()));
						te.setAttribute("line", String.valueOf(token.getLine()));
						te.setAttribute("type", String.valueOf(repoLanguage.tokenType2id(token.getType())));
						if (token.getText()!=null) {
							te.appendChild(doc.createTextNode(token.getText()));
						}
					}

					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					GZIPOutputStream gzos = new GZIPOutputStream(baos);
			        TransformerFactory.newInstance().newTransformer().transform(new DOMSource(doc), new StreamResult(gzos));
					gzos.close();
					lli.setLoaderData(baos.toByteArray());
					
					return true;
				} finally {
					r.close();
				}
			} catch (Exception e) {
				lli.setLoadFailed(true);
				lli.setMessageId(storeErrorMessage(scanId, sourceUnitId, e));
				logger.log(Level.WARNING, "Cannot load source '"+source+"' to level "+data.getLevel()+": "+e, e);
				return false;
			} finally {
				factory.getEngine().insertLoadLevel(lli);
			}
		} catch (SQLException e) {
			factory.consume(this, e);
			return false;
		}
	}
	
	/**
	 * Reads tokens from BLOB and returns collection of biz.hammurapi.antlr.Token instances
	 */
	public Object getData(int sourceUnitId, Integer scanId) throws MesopotamiaException {
		RepositoryFactory factory=repoLanguage.getFactory();
		try {
			LoadLevel loadLevel = factory.getEngine().getSourceUnitLoadLevel(data.getId(), sourceUnitId);
			if (loadLevel==null) {
				return null;
			}
			Collection<MesopotamiaToken> ret = new ArrayList<MesopotamiaToken>();						
		    Document doc=DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(new GZIPInputStream(new ByteArrayInputStream(loadLevel.getLoaderData())));
		    NodeList nl = doc.getDocumentElement().getChildNodes();
		    MesopotamiaToken prevToken = null;
			for (int i = 0, l = nl.getLength(); i<l; ++i) {
				Node n = nl.item(i);
				if (n instanceof Element) {
					Element el = (Element) n;
					int typeId = Integer.parseInt(el.getAttribute("type"));
					MesopotamiaToken token = new MesopotamiaToken(i, repoLanguage.isWhitespace(typeId));  
					token.setType(repoLanguage.tokenTypeId2type(typeId));
					token.setTypeName(repoLanguage.tokenTypeId2name(typeId));
					token.setText(AbstractDomObject.getElementText(el));
					token.setColumn(Integer.parseInt(el.getAttribute("col")));
					token.setLine(Integer.parseInt(el.getAttribute("line")));
					token.setFilename("Source unit "+sourceUnitId); // TODO - real name.
					token.setPrevToken(prevToken);
					ret.add(token);
					prevToken = token;					
				}
			}

			return ret;
		} catch (Exception e) {
			throw new MesopotamiaException("Cannot read tokens: "+e, e);
		}
	}
}
