/*
  * mesopotamia-java @mesopotamia.version@
 * Multilingual parser and repository. 
 * Copyright (C) 2005  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://http://www.hammurapi.biz
 * e-Mail: support@hammurapi.biz
 */
package org.mesopotamia.lang.java.v5;

import gnu.trove.TIntIntHashMap;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.mesopotamia.Loader;
import org.mesopotamia.LoaderBase;
import org.mesopotamia.LoaderEntry;
import org.mesopotamia.MesopotamiaException;
import org.mesopotamia.MesopotamiaToken;
import org.mesopotamia.RepositoryFactory;
import org.mesopotamia.RepositoryLanguage;
import org.mesopotamia.SourceUnitLoader;
import org.mesopotamia.SyntaxTree;
import org.mesopotamia.lang.java.JavaDocComment;
import org.mesopotamia.lang.java.MesopotamiaJavaNode;
import org.mesopotamia.lang.java.MultiLineComment;
import org.mesopotamia.lang.java.SingleLineComment;
import org.mesopotamia.sql.LoadLevel;
import org.mesopotamia.sql.LoadLevelImpl;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

import antlr.ASTFactory;
import antlr.Token;
import antlr.TokenStream;
import antlr.TokenStreamException;
import antlr.collections.AST;
import biz.hammurapi.antlr.AstVisualizable;
import biz.hammurapi.swing.Browser;
import biz.hammurapi.xml.dom.AbstractDomObject;

public class AstLoader extends LoaderBase implements SourceUnitLoader {
	private static final String JAVA_DOC_START = "/**";
	private static final String ELEMENT_NODE = "node";
	private static final String ELEMENT_MULTI_LINE_COMMENT = "multi-line-comment";
	private static final String ELEMENT_SINGLE_LINE_COMMENT = "single-line-comment";
	private static final String ATTRIBUTE_TEXT = "text";
	private static final String ATTRIBUTE_LAST_TOKEN = "last-token";
	private static final String ATTRIBUTE_FIRST_TOKEN = "first-token";
	private static final String ATTRIBUTE_TOKEN = "token";
	private static final String ATTRIBUTE_SAME_TYPE_INDEX = "same-type-index";
	private static final String ATTRIBUTE_TYPE = "type";
	private static final String ATTRIBUTE_ID = "id";
	private static final String ATTRIBUTE_LEFT_COLUMN = "left-column";
	private static final String ATTRIBUTE_LINE = "line";
	private static final String ATTRIBUTE_COL = "col";
	private static final Logger logger = Logger.getLogger(AstLoader.class.getName());
	private LoaderEntry xData;

	public AstLoader(RepositoryLanguage repoLanguage, LoaderEntry xData) {
		super(repoLanguage, xData);
		this.xData = xData;
	}

	public boolean load(int scanId, final int sourceUnitId, Object environment) {
		final RepositoryFactory factory = repoLanguage.getFactory();
		//final SourceUnit si = (SourceUnit) sit.next();
		final String sourceUnitName = "Source unit #"+sourceUnitId;
		logger.fine("Loading source unit '"+sourceUnitName+"' to level "+xData.getLevel());

		try {
			LoadLevelImpl lli = new LoadLevelImpl(true);
			lli.setLevelId(xData.getId());
			lli.setSourceUnitId(sourceUnitId);
			try {
				Loader tokenLoader = repoLanguage.getLoader(ATTRIBUTE_TOKEN);				
				Collection<Token> tokens = (Collection<Token>) tokenLoader.getData(sourceUnitId, new Integer(scanId));
				Collection<Token> nonWhiteSpace = null;
				if (tokens!=null) {
					nonWhiteSpace = new ArrayList<Token>(tokens);
					Iterator<Token> it = nonWhiteSpace.iterator();
					while (it.hasNext()) {
						Object t = it.next();
						if (t instanceof MesopotamiaToken && ((MesopotamiaToken) t).isWhitespace()) {
							it.remove();
						}						
					}
				}
				
				if (nonWhiteSpace!=null && !nonWhiteSpace.isEmpty()) {
					nonWhiteSpace.add(new Token(JavaTokenTypes.EOF));
					final Iterator<Token> tit = nonWhiteSpace.iterator();
					JavaRecognizer parser = new JavaRecognizer(new TokenStream() {
	
						public Token nextToken() throws TokenStreamException {
							return (Token) (tit.hasNext() ? tit.next() : null);
						}
	
					});
	
					ASTFactory astFactory = new ASTFactory();
					astFactory.setASTNodeClass(biz.hammurapi.antlr.AST.class);
					parser.setASTFactory(astFactory);
	
					parser.compilationUnit();
					TIntIntHashMap sameTypeIndices = new TIntIntHashMap();
					AST compilationUnitAst = parser.getAST();					
//					Browser.show(new AstVisualizable(null, compilationUnitAst, parser.getTokenNames(), true), "CU");
//					((biz.hammurapi.antlr.AST) compilationUnitAst).showWithSiblings("Compilation unit", parser.getTokenNames());
					
				    Document doc=DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
				    Element rootElement = AbstractDomObject.addElement(doc, "syntax-tree");
				    int[] counter = {0};
					
				    Set<Integer> usedComments = new HashSet<Integer>();
					for (AST node = compilationUnitAst; node != null; node = node.getNextSibling()) {
						int sameTypeIndex = sameTypeIndices.get(node.getType());
						sameTypeIndices.put(node.getType(), sameTypeIndex + 1);
						((biz.hammurapi.antlr.AST) node).queryChildren();						
						store(rootElement, (biz.hammurapi.antlr.AST) node, sameTypeIndex, counter, usedComments);
					}

					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					GZIPOutputStream gzos = new GZIPOutputStream(baos);
			        TransformerFactory.newInstance().newTransformer().transform(new DOMSource(doc), new StreamResult(gzos));
//			        TransformerFactory.newInstance().newTransformer().transform(new DOMSource(doc), new StreamResult(System.out));
					gzos.close();
					lli.setLoaderData(baos.toByteArray());					
				}
				return true;
			} catch (Exception e) {
				lli.setLoadFailed(true);
				lli.setMessageId(storeErrorMessage(scanId, sourceUnitId, e));
				logger.log(Level.WARNING, "Cannot load source unit '"+sourceUnitName+"' to level "+xData.getLevel()+": "+e, e);
				return false;
			} finally {
				factory.getEngine().insertLoadLevel(lli);
			}
		} catch (SQLException e) {
			factory.consume(this, e);
			return false;
		}
	}

	private void store(Element owner, biz.hammurapi.antlr.AST ast, int sameTypeIndex, int[] counter, Set<Integer> usedComments) throws SQLException {
		Element holder = owner.getOwnerDocument().createElement(ELEMENT_NODE);
		owner.appendChild(holder);
		
		holder.setAttribute(ATTRIBUTE_COL, String.valueOf(ast.getColumn()));
		holder.setAttribute(ATTRIBUTE_LINE, String.valueOf(ast.getLine()));
		holder.setAttribute(ATTRIBUTE_LEFT_COLUMN, String.valueOf(ast.getLeftColumn()));
		holder.setAttribute(ATTRIBUTE_ID, String.valueOf(counter[0]++));
		holder.setAttribute(ATTRIBUTE_TYPE, String.valueOf(repoLanguage.tokenType2id(ast.getType())));
		holder.setAttribute(ATTRIBUTE_SAME_TYPE_INDEX, String.valueOf(sameTypeIndex));
		
		if (ast.getToken()!=null) {
			holder.setAttribute(ATTRIBUTE_TOKEN, String.valueOf(((MesopotamiaToken) ast.getToken()).getId()));
		}
		if (ast.getFirstToken()!=null) {
			holder.setAttribute(ATTRIBUTE_FIRST_TOKEN, String.valueOf(((MesopotamiaToken) ast.getFirstToken()).getId()));
		}
		if (ast.getLastToken()!=null) {
			holder.setAttribute(ATTRIBUTE_LAST_TOKEN, String.valueOf(((MesopotamiaToken) ast.getLastToken()).getId()));
		}
		if (ast.getText()!=null) {
			holder.setAttribute(ATTRIBUTE_TEXT, ast.getText());
		}
				
		if (ast.getType()!=JavaTokenTypes.OBJBLOCK) {
			LinkedList<MesopotamiaToken> comments = new LinkedList<MesopotamiaToken>();
			MesopotamiaToken firstToken = (MesopotamiaToken) ast.getFirstToken();
			MesopotamiaToken wsToken = (MesopotamiaToken) (firstToken==null ? null : firstToken.getPrevToken());
			while (wsToken!=null && wsToken.isWhitespace()) {
				if (wsToken.getType()==JavaTokenTypes.SL_COMMENT || wsToken.getType()==JavaTokenTypes.ML_COMMENT) {
					if (usedComments.add(wsToken.getId())) {
						comments.addFirst(wsToken);
					}
				}
				wsToken = (MesopotamiaToken) wsToken.getPrevToken();
			}
			
			for (MesopotamiaToken comment: comments) {
				Element ce;
				if (comment.getType()==JavaTokenTypes.SL_COMMENT) {
					ce = AbstractDomObject.addTextElement(holder, ELEMENT_SINGLE_LINE_COMMENT, comment.getText());
				} else {
					ce = AbstractDomObject.addTextElement(holder, ELEMENT_MULTI_LINE_COMMENT, comment.getText());				
				}
				ce.setAttribute(ATTRIBUTE_LINE, String.valueOf(comment.getLine()));
				ce.setAttribute(ATTRIBUTE_COL, String.valueOf(comment.getColumn()));			
			}
		}
		
		TIntIntHashMap sameTypeIndices = new TIntIntHashMap();
		int childPosition = 0;
		for (AST child = ast.getFirstChild(); child != null; child = child.getNextSibling(), ++childPosition) {
			int childSameTypeIndex = sameTypeIndices.get(child.getType());
			sameTypeIndices.put(child.getType(), childSameTypeIndex + 1);
			store(holder, (biz.hammurapi.antlr.AST) child, childSameTypeIndex, counter, usedComments);
		}
	}

	/**
	 * Returns SyntaxTree
	 */
	public Object getData(final int sourceUnitId, Integer scanId) throws MesopotamiaException {
		RepositoryFactory factory=repoLanguage.getFactory();
		try {
			LoadLevel loadLevel = factory.getEngine().getSourceUnitLoadLevel(data.getId(), sourceUnitId);
			if (loadLevel==null) {
				return null;
			}
			
			byte[] loaderData = loadLevel.getLoaderData();
			if (loaderData==null) {
				return null;
			}

			javax.xml.parsers.SAXParserFactory saxParserFactory = javax.xml.parsers.SAXParserFactory.newInstance();     
			javax.xml.parsers.SAXParser  saxParser = saxParserFactory.newSAXParser();
			XMLReader parser = saxParser.getXMLReader();
			final SyntaxTree ret = new SyntaxTree();
			ContentHandler ch = new DefaultHandler() {
				MesopotamiaJavaNode currentNode;
				StringBuilder currentComment;
				int currentCommentLine;
				int currentCommentColumn;
				
				/**
				 * Writes text to node, nullifies currentText.
				 */
				public void endElement(String uri, String localName, String qName) throws SAXException {
					if (ELEMENT_NODE.equals(qName)) {
						if (currentNode!=null) {
							currentNode=(MesopotamiaJavaNode) currentNode.getParent();
						}
					} else if (ELEMENT_MULTI_LINE_COMMENT.equals(qName)) {
						String text = currentComment.toString();
						if (text!=null && text.startsWith(JAVA_DOC_START)) {
							currentNode.addComment(new JavaDocComment(currentCommentLine, currentCommentColumn, text));							
						} else {
							currentNode.addComment(new MultiLineComment(currentCommentLine, currentCommentColumn, text));														
						}
						currentComment = null;
					} else if (ELEMENT_SINGLE_LINE_COMMENT.equals(qName)) {
						currentNode.addComment(new SingleLineComment(currentCommentLine, currentCommentColumn, currentComment.toString()));
						currentComment = null;
					}
				}
				
				public void error(SAXParseException e) throws SAXException {
					throw e;
				}
				
				public void fatalError(SAXParseException e) throws SAXException {
					throw e;
				}
				
				public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
					if (ELEMENT_NODE.equals(qName)) {
						int tokenType = Integer.parseInt(attributes.getValue(ATTRIBUTE_TYPE));
						MesopotamiaJavaNode newNode = new MesopotamiaJavaNode(repoLanguage.tokenTypeId2name(tokenType), currentNode);
						
						newNode.setColumn(Integer.parseInt(attributes.getValue(ATTRIBUTE_COL)));
						newNode.setLine(Integer.parseInt(attributes.getValue(ATTRIBUTE_LINE)));
						newNode.setType(tokenType);
						newNode.setSameTypeIndex(Integer.parseInt(attributes.getValue(ATTRIBUTE_SAME_TYPE_INDEX)));
						newNode.setLeftColumn(Integer.parseInt(attributes.getValue(ATTRIBUTE_LEFT_COLUMN)));
						newNode.setId(Integer.parseInt(attributes.getValue(ATTRIBUTE_ID)));
						newNode.setText(attributes.getValue(ATTRIBUTE_TEXT));
						newNode.setSourceUnitId(sourceUnitId);
						ret.mapNode(newNode);
						
						
						String ft = attributes.getValue(ATTRIBUTE_FIRST_TOKEN);
						if (ft!=null) {
							newNode.setFirstToken(new Integer(ft));
						}
						
						String lt = attributes.getValue(ATTRIBUTE_LAST_TOKEN);
						if (lt!=null) {
							newNode.setLastToken(new Integer(lt));
						}
						
						String t = attributes.getValue(ATTRIBUTE_TOKEN);
						if (t!=null) {
							newNode.setToken(new Integer(t));
						}
						
						if (currentNode==null) {
							newNode.setPosition(ret.getRoots().size());
							ret.getRoots().add(newNode);
						} else {
							newNode.setPosition(currentNode.getChildren().size());
							currentNode.getChildren().add(newNode);
						}
						currentNode=newNode;
					} else if (ELEMENT_MULTI_LINE_COMMENT.equals(qName) || ELEMENT_SINGLE_LINE_COMMENT.equals(qName)) {
						currentComment = new StringBuilder();
						currentCommentLine = Integer.parseInt(attributes.getValue(ATTRIBUTE_LINE));
						currentCommentColumn = Integer.parseInt(attributes.getValue(ATTRIBUTE_COL));
					}
				}	
				
				@Override
				public void characters(char[] ch, int start, int length) throws SAXException {
					if (currentComment==null) {
						super.characters(ch, start, length);
					} else {
						currentComment.append(ch, start, length);
					}
				}
				
			};
			parser.setContentHandler(ch);			
			parser.parse( new org.xml.sax.InputSource(new GZIPInputStream(new ByteArrayInputStream(loaderData))));			
			return ret;
		} catch (Exception e) {
			throw new MesopotamiaException("Cannot load AST: "+e, e);
		}
	}
}
